#! /usr/bin/env python2.1

#######################################################
#
#    Winzig's Database Engine
#
#    Copyright (C) 2001,2002 Robert Muth <robert at muth dot org>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; version 2 of June 1991.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program in form of the file COPYING; 
#    if not, write to the 
#
#    Free Software Foundation, Inc. <http://www.fsf.org>
#    59 Temple Place, Suite 330, 
#    Boston, MA 02111-1307  USA
#    $Id: wdatabase.py,v 1.5 2003/12/04 22:37:20 muth Exp $
#
#######################################################

from __future__ import nested_scopes

import sys
import os
import getopt
import re
import string
import copy

#######################################################################
# Version
#######################################################################

def Version():
    
    (l,v,x) = string.split('$Revision: 1.5 $')
    return v

#######################################################################
# Usage
#######################################################################

def Usage():
    print "Usage"

#######################################################################
# Exceptions
#######################################################################

class DatabaseError( Exception ):
    def __init__(self,message):
        self._message = message

    def message(self):
        return self._message

#######################################################################
# Record Structures
#######################################################################

class ClassRecord:
    def __init__(self):
        self._fields = {}

    def init_from_raw(self,pairs,line):
        for (tag,value) in pairs:
            if self._fields.has_key(tag):
                raise DatabaseError("Duplicate tag: " + tag + " " + value + (" in line %d" % line))
            self._fields[tag] = value

    def update_field(self,tag,value):
        self._fields[tag] = value

    def get_field(self,tag):
        return self._fields.get(tag,"")
        
    def tags(self):
        return self._fields.keys()

    def write(self,output,tags):
        for tag in tags:
            value = self._fields.get(tag,None)
            if value:
                output.write(tag + " " + value + "\n")

#######################################################################
#
#######################################################################

sort_tag = ""

def my_sorter(a,b):
    global sort_tag
    x = string.upper(a.get(sort_tag,""))
    y = string.upper(b.get(sort_tag,""))
    if x < y:
        return -1
    elif y < x:
        return 1
    else:
        return 0


class ClassAttribute:
      
    def __init__(self,tag,number):
        self._tag     = tag
        self._number = number
         
        self._name  = ""
        self._required  =  0
        self._static  = 0
        self._list = None
        self._value = None

    def init_from_raw(self,properties):
        for  (p,v) in properties:
            if p == "name": self._name = v
            elif p == "static": self._static = 1
            elif p == "required": self._required = 1
            elif p == "value": self._value = v
            elif p == "list": self._list = string.split(v,",")
            else:
                raise DatabaseError("unknown attribute property: " + p) 
                     
        if self.name == "":
            raise DatabaseError("no name for attribute " + tag)

    def tag(self):
        return self._tag

    def number(self):
        return self._number

    def static(self):
        return self._static

    def name(self):
        return self._name

    def value(self):
        return self._value

    def list(self):
        list = self._list
        if list: list = list[:]
        return list

    def required(self):
        return self._required

    def write(self,output):
        output.write(self._tag)
        output.write(" name(" + self._name + ")")
        if self._required:
            output.write(" required()")
        if self._static:
            output.write(" static()")
        if self._value:
            output.write(" value(" + self._value + ")")
        if self._list:
            output.write(" list(" + string.join(self._list,",") + ")")
        output.write("\n")
        return
    
#######################################################################
#
#######################################################################

class ClassSchema:
    def __init__(self):
        self._attributes = {}

    def get_attribute(self,tag):
        return self._attributes.get(tag,None)

    def get_all_attributes(self,tag):
        return self._attributes.values()

    def get_all_attributes_tags(self):
        return self._attributes.keys()

    def get_attribute_tags_original_order(self):
        def cmp(a1,a2):
            return self._attributes[a1].number() - self._attributes[a2].number()

        tags = self._attributes.keys()
        tags.sort( cmp )
        return tags
    
    patProperty = re.compile('(\w+)\((.*?)\)')

    def init_from_raw(self,pairs,line):
        count = 0
        for (tag,value) in pairs:
            matches = ClassSchema.patProperty.findall(value)
            properties = []
            for tuple in matches:
                properties.append(tuple)
            attribute = ClassAttribute(tag,count)
            attribute.init_from_raw(properties)
            if self._attributes.has_key(tag):
                raise DatabaseError("duplicate attribute: " + tag)
            self._attributes[tag] = attribute
            count = count + 1

    def check_record(self,record,line):
        if line:
            line_info =  " line %d" % line
        else:
            line_info = ""
        for attribute in self._attributes.values():
            field = record.get_field(attribute.tag())
            if attribute.static():
                continue
            if attribute.required() and field == "":
                raise DatabaseError("missing required attribute\n " + attribute.tag() + ": " + attribute.name() + line_info)       
            if attribute.list() and field != "" and not field in attribute.list():
                raise DatabaseError(field + " not in list for " + attribute.tag() + line_info)       
            
        for tag in record.tags():
            attribute = self.get_attribute(tag)
            if not attribute:
                raise DatabaseError("Unknown tag: " + tag )                  
            if attribute.static():
                raise DatabaseError("static tag encountered: " + tag )
        return
     
    def write(self,output):
        tags = self.get_attribute_tags_original_order()
        for tag in tags:
            self._attributes[tag].write(output)

#######################################################################
#
#######################################################################

def SaveBackups(filename,num):
      for i in range(num,0,-1):
            src = filename + ".%d" % i
            dst = filename + ".%d" % (i+1)
            #print "moving",src,dst
            try:
                  os.rename( src, dst)
            except:
                  pass
      os.rename(filename, filename + ".1")

#######################################################################
#
#######################################################################
                  
class ClassDatabase:

    def __init__(self):
        self._filename = ""
        self._dirty = 0
        self._schema = None
        self._records = []

    def num_records(self):
        return len(self._records)

    def set_dirty(self,state = 1):
        self._dirty = state

    def update_record(self,id,record):
        self._schema.check_record(record,None)
        self._records[id] = copy.deepcopy(record)
        self.set_dirty()
        
    def add_record(self,record,line=None):
        self._schema.check_record(record,line)
        self._records.append( copy.deepcopy(record))
        self.set_dirty()
        return len(self._records) - 1
    
    def del_record(self,id):
        self._records[id] = None
        self.set_dirty()

    def get_record(self,id):
        return copy.deepcopy(self._records[id])

    def get_record_ids(self, test):
        if not test: test = lambda y: 1
        return filter( lambda x: self._records[x] and test(self._records[x]), range( len(self._records)))

    def get_attribute(self,tag):
        return self._schema.get_attribute(tag)
    
    def read_from_file(self,filename):
        self._filename = filename
        
        patComment = re.compile('^#.*$')
        patEmpty = re.compile('^[ \t]*$')
        patTagValue = re.compile('^(\w\w)[\t ]+(\S.*)$') 

        try:
            input = open(filename, "r")
            lines = input.readlines()
            input.close()
        except:
            raise DatabaseError("cannot read database " + filename)

        #
        # read set of pairs
        # first set of pairs is scheme
        # remaining set of pairs pairs are records
        #
        
        line_number = 0
        pairs = []
        for line in lines:
        
            line_number = line_number + 1

            # skip comments
            match = patComment.search(line)
            if match: continue

            # an empty line represents a record terminator
            
            match = patEmpty.search(line)
            if match: 
                if pairs != []:
                    if self._schema:
                          record = ClassRecord()
                          record.init_from_raw(pairs,line_number)
                          self.add_record(record,line_number)                          
                    else:
                          self._schema = ClassSchema()
                          self._schema.init_from_raw(pairs,line_number)
                    pairs = []
                continue

            match = patTagValue.search(line)
            if match:
                tag =  match.group(1)
                value = match.group(2)
                pairs.append( (tag,value) )
                continue

            raise DatabaseError("Format error: " + "%6d: " % line_number + line[:-1])
        
        self.set_dirty(0)

    def write_prolog(self,output):
        output.write("#\n")
        output.write("# Winzig Database\n")
        output.write("# Edit with care!\n")
        output.write("# Use utf-8 encoding!\n")
        output.write("#\n")

    def write_epilog(self,output):
        output.write("#\n")
        output.write("#eof\n")
        output.write("#\n")
        
    def write_to_file(self, sort=1, filename = None):
        if filename == None:
            filename = self._filename

        SaveBackups(filename,8)
        output = open(filename, "w")
        self.write_prolog(output)
        self._schema.write(output)
        output.write("\n")

        sort_tags = filter( lambda x: not self.get_attribute(x).static(), self._schema.get_attribute_tags_original_order())

        sorted_records =  filter( lambda x: x, self._records) # this is not a deep copy!
        
        if sort:
            def cmp(r1,r2):
                for tag in sort_tags:
                    s1 = r1.get_field(tag)
                    s2 = r2.get_field(tag)
                    if s1 == s2: continue
                    if s1 < s2: return -1
                    else: return +1
                return 0
            sorted_records.sort( cmp )

        for record in sorted_records:
            record.write(output,sort_tags)
            output.write("\n")

        self.write_epilog(output)
    
        output.close()          

#######################################################################
# Main
#######################################################################

def Main(argv):

    #
    # SETTING DEFAULTS
    #
    
    Config = {}
    Config['version'] = Version()
    Config['database'] = 'contact.db'

    #
    # COMMAND LINE PARSING & ACTION
    # 
    
    try:
        mode = argv[1]
        opts, args = getopt.getopt(argv[2:], "d:")
        for o, a in opts:
            if o == '-d':
                Config['database'] =  a
                continue

        if mode == 'test':
            db = ClassDatabase()
            db.read_from_file(Config['database'])
            sys.exit(0)

        elif mode == 'query':
            db = ClassDatabase()
            db.read_from_file(Config['database'])

            sys.exit(0)                                  
        else:
            Usage()
            sys.exit(-1)                                  

    except DatabaseError,e:
        print e.message()
        sys.exit(-1)                          

#    except:
#        Usage() 
#        sys.exit(-1)                          

    sys.exit(0)                          
#######################################################
#
#######################################################

if __name__ == "__main__": Main( sys.argv)

#######################################################
# eof
#######################################################




